unit BL_Flash;

{$IFDEF NoDebug}

{$O+} // Optimisation ON
{$D-} // Debug information OFF
{$I-} // I/O checking OFF
{$L-} // Local Symbols OFF
{$Q-} // Overflow Checking OFF
{$R-} // Range Checking OFF

{$ELSE}
{$O-} // Optimisation OFF
{$D+} // Debug information ON
{$I+} // I/O checking ON
{$L+} // Local Symbols ON
{$Q+} // Overflow Checking ON
{$R+} // Range Checking ON

{$ENDIF}

interface

procedure testboot;

procedure CheckLoader; stdcall;

function ParseFile(FName: string): Boolean;
procedure LoadApplication(FName: string);

implementation
uses
  Forms,
  Math,
  HIDinterface,
  SysUtils,
  crd_Time,
  BL_Main,
  MG_IO,
  BL_Vars;

const
  Reset_Vector_size_P16 = 2;
  Reset_Vector_Size_P18 = 4;
  Reset_Vector_Size_P24 = 6;
  reset_Vector_Size_P32 = 16;

function Bytes_to_Addresses(Bytes: int64): int64;
begin
  case TargetDevice of
    mtPIC16: Result := Bytes div 2;
    mtPIC18, mtPIC18FJ: Result := Bytes;
    mtPIC24, mtdsPIC: Result := Bytes * 2 div 3;
    mtPIC32: Result := Bytes;
  else Result := Bytes;
  end
end;

function Addresses_to_Bytes(Addresses: int64): int64;
begin
  case TargetDevice of
    mtPIC16: Result := Addresses * 2;
    mtPIC18, mtPIC18FJ: Result := Addresses;
    mtPIC24, mtdsPIC: Result := Addresses * 3 div 2;
    mtPIC32: Result := Addresses;
  else Result := Addresses;
  end
end;

procedure ShowBootInfo(Data: TBootInfo);
begin
  with Data do
  begin
    case bMCUtype.fValue of
      mtPIC16: ShowInfo('   MCUtype      : ' + stPIC16);
      mtPIC18: ShowInfo('   MCUtype      : ' + stPIC18);
      mtPIC18FJ: ShowInfo('   MCUtype      : ' + stPIC18FJ);
      mtPIC24: ShowInfo('   MCUtype      : ' + stPIC24);
      mtDSPIC: ShowInfo('   MCUtype      : ' + stDSPIC);
      mtPIC32: ShowInfo('   MCUtype      : ' + stPIC32);
    else ShowInfo('   MCUtype      : ' + IntToHex(bMCUtype.fValue, 2));
    end;

    case TargetDevice of
      mtPIC16, mtPIC18, mtPIC18FJ:
        begin
          ShowInfo('   MCU size     : 0x' + IntToHex(Bytes_to_Addresses(ulMcuSize.fValue), 8) + ' => ' + IntToStr(ulMcuSize.fValue) + ' bytes');
          ShowInfo('   Erase block  : 0x' + IntToHex(Bytes_to_Addresses(uiEraseBlock.intval), 4) + '     => ' + IntToStr(uiEraseBlock.intVal) + ' bytes');
          ShowInfo('   Write block  : 0x' + IntToHex(Bytes_to_Addresses(uiWriteBlock.intVal), 4) + '     => ' + IntToStr(uiWriteBlock.intVal) + ' bytes');
          ShowInfo('   Boot start   : 0x' + IntToHex(ulBootStart.fValue, 8) + ' => ' + IntToStr(ulBootStart.fValue));
          ShowInfo('   Boot revision: 0x' + IntToHex(uiBootRev.intVal, 4));
        end;

      mtPIC24, mtdsPIC:
        begin
          ShowInfo('   MCU size     : 0x' + IntToHex(Bytes_to_Addresses(ulMcuSize.fValue), 8) + ' Addresses => ' + IntToStr(ulMcuSize.fValue) + ' bytes');
          ShowInfo('   Erase block  : 0x' + IntToHex(Bytes_to_Addresses(uiEraseBlock.intval), 4) + ' Addresses     => ' + IntToStr(uiEraseBlock.intVal) + ' bytes');
          ShowInfo('   Write block  : 0x' + IntToHex(Bytes_to_Addresses(uiWriteBlock.intVal), 4) + ' Addresses     => ' + IntToStr(uiWriteBlock.intVal) + ' bytes');
          ShowInfo('   Boot start   : 0x' + IntToHex((ulBootStart.fValue), 8) + ' Addresses => ' + IntToStr(Addresses_to_bytes(ulBootStart.fValue)) + ' bytes');
          ShowInfo('   Boot revision: 0x' + IntToHex(uiBootRev.intVal, 4));
        end;

      mtPIC32:
        begin
          ShowInfo('   MCU size     : 0x' + IntToHex(Bytes_to_Addresses(ulMcuSize.fValue), 8) + ' => ' + IntToStr(ulMcuSize.fValue) + ' bytes');
          ShowInfo('   Erase block  : 0x' + IntToHex(Bytes_to_Addresses(uiEraseBlock.intval), 4) + '     => ' + IntToStr(uiEraseBlock.intVal) + ' bytes');
          ShowInfo('   Write block  : 0x' + IntToHex(Bytes_to_Addresses(uiWriteBlock.intVal), 4) + '     => ' + IntToStr(uiWriteBlock.intVal) + ' bytes');
          ShowInfo('   Boot start   : 0x' + IntToHex(ulBootStart.fValue, 8) + ' => ' + IntToStr(BootStartZeroBased));
          ShowInfo('   Boot revision: 0x' + IntToHex(uiBootRev.intVal, 4));
        end;
    end;

    ShowInfo('   Descriptor   : ' + sDevDsc.fValue);
  end;
end;

function UpperBlockBoundary(Value, BlockSize: longword): longword;
var
  Remainder: word;
begin
  Result := Value;
  Remainder := Value mod BlockSize;
  if Remainder > 0 then
    inc(Result, BlockSize - Remainder);
end;

function LowerBlockBoundary(Value, BlockSize: longword): longword;
var
  Remainder: word;
begin
  Result := value;
  Remainder := Value mod BlockSize;
  dec(Result, Remainder);
end;

procedure LoadBootRecord(InData: THIDbuffer;
  var OutData: TBootInfo);
// load boot record according to mE definitions
(*
type TBootInfo = record
  bSize       : byte;         // value only
  bMcuType    : TCharField;   // label + 1 byte value
  ulMcuSize   : TULongField;  // label + 4 byte value
  uiEraseBlock: TUIntField;   // label + 2 byte value
  uiWriteBlock: TUIntField;   // label + 2 byte value
  uiBootRev   : TUIntField;   // label + 2 byte value
  ulBootStart : TULongField;  // label + 4 byte value
  sDevDsc     : TStringField; // label + string (end is bsize)
end;
*)

var
  Index, WordAlignment, DWordAlignment: Integer;
begin

  with OutData do
  begin
    WordAlignment := 1; // default
    DWordAlignment := 1;

    Index := 0;
    bsize := Indata[Index];

    inc(Index);
    repeat
      case Indata[Index] of
        bifMCUTYPE:
          begin
            inc(Index); // point to value

            bMCUtype.fFieldType := bifMCUTYPE;
            bMCUtype.fValue := InData[Index];
            TargetDevice := bMCUtype.fValue;

            case TargetDevice of
              mtPIC16, mtPIC18, mtPIC18FJ:
                begin
                  WordAlignment := 1;
                  DWordAlignment := 1;
                end;
              mtPIC24, mtDsPIC:
                begin
                  WordAlignment := 2;
                  DWordAlignment := 2;
                end;
              mtPIC32:
                begin
                  WordAlignment := 2;
                  DWordAlignment := 4;
                end;
            end;

            inc(Index); Index := UpperBlockBoundary(Index, DwordAlignment); // point to next item
          end;

        bifMCUSIZE:
          begin
            inc(Index); Index := UpperBlockBoundary(Index, DWordAlignment); // point to value

            ulMcuSize.fFieldType := bifMCUSIZE;
            ulMcuSize.fValue := (InData[Index + 3] shl 24) + (InData[Index + 2] shl 16) +
              (InData[Index + 1] shl 8) + InData[Index];

            inc(Index, 4); Index := UpperBlockBoundary(Index, DWordAlignment); // point to next item
          end;

        bifERASEBLOCK:
          begin
            inc(Index); Index := UpperBlockBoundary(Index, WordAlignment); // point to value

            uiEraseBlock.fFieldType := bifERASEBLOCK;
            uiEraseBlock.intVal := (InData[Index + 1] shl 8) + InData[Index];

            inc(Index, 2); Index := UpperBlockBoundary(Index, DWordAlignment); // point to next item
          end;

        bifWriteBLOCK:
          begin
            inc(Index); Index := UpperBlockBoundary(Index, WordAlignment); // point to value

            uiWriteBlock.fFieldType := bifWRITEBLOCK;
            uiWriteBlock.intVal := (InData[Index + 1] shl 8) + InData[Index];

            inc(Index, 2); Index := UpperBlockBoundary(Index, DWordAlignment); // point to next item
          end;

        bifBOOTSTART:
          begin
            inc(Index); Index := UpperBlockBoundary(Index, DWordAlignment); // point to value

            ulBootStart.fFieldType := bifBOOTSTART;
            ulBootStart.fValue := (InData[Index + 3] shl 24) + (InData[Index + 2] shl 16) +
              (InData[Index + 1] shl 8) + InData[Index];

            BootStartZeroBased := ulBootStart.fValue and $00FFFFFF;

            inc(Index, 4); Index := UpperBlockBoundary(Index, DWordAlignment); // point to next item
          end;

        bifBOOTREV:
          begin
            inc(Index); Index := UpperBlockBoundary(Index, WordAlignment); // point to value

            uiBootRev.fFieldType := bifBOOTREV;
            uiBootRev.intVal := (InData[Index + 1] shl 8) + InData[Index];

            inc(Index, 2); Index := UpperBlockBoundary(Index, DWordAlignment); // point to next item
          end;

        bifDEVDSC:
          begin
            inc(Index); // point to value

            sDevDsc.fFieldType := bifDEVDSC;
            sDevDsc.fValue := '';
            while Index < bsize - 1 do
            begin
              sDevDsc.fValue := sDevDsc.fValue + CHR(InData[Index]);
              inc(Index);
            end;
          end

      else
        begin
          inc(Index); Index := UpperBlockBoundary(Index, DWordAlignment); // point to next item
        end;
      end;
    until Index >= bsize - 1;
  end;

end;

procedure SendCommand_GetReply(Cmd: Byte;
  var Data: THIDbuffer);
// send command to device and get reply
var
  Start: Integer;
  Ok: Boolean;
begin
  if (CurrentDevice = -1) then Exit;
  for Start := 0 to 63 do Data[Start] := 0;
  Data[0] := STX; // STX command
  Data[1] := Cmd; // cmd opcode
  HIDwriteDevice(ActiveDevice, Data); // send request to active device
  Sleep(10);

  for Start := 0 to 63 do Data[Start] := 0;
  Start := 0; // reset time out timer
  repeat
    OK := HIDreadDevice(ActiveDevice, Data); // try to get response from target device
    if not OK then
    begin
      Sleep(12); // <-----------
      INC(Start);
    end;
  until OK or // response received
    (Start > 500); // time out // <-----------

  if OK then // response received
  else
  begin
    inc(numerrors);
    if (Cmd <> $04) then ErrorCode := ErrorCode or $20;
  end;
end;

procedure SendCommandNoAck(Cmd: Byte;
  var Data: THIDbuffer);
// send command to device
begin
  if (CurrentDevice = -1) then Exit;
  Data[0] := STX; // STX command
  Data[1] := Cmd; // cmd opcode
  HIDwriteDevice(ActiveDevice, Data); // send request to active device
end;

function SendDataNoAck(Data: THIDbuffer): Boolean;
begin
  Result := FALSE;
  if (NumErrors > 0) then Exit;
  if (CurrentDevice = -1) then Exit;

  HIDwriteDevice(DevList[CurrentDevice], Data); // send request to active device
  Sleep(12);

  Result := TRUE;
end;

function SendDataWithAck(Data: THIDbuffer; ACK: Byte): Boolean;
// send data to device and wait for acknowledge
var
  Start: Integer;
  OK: Boolean;
  Buffer: THIDbuffer;
begin
  Result := FALSE;
  if (NumErrors > 0) then Exit;
  if (CurrentDevice = -1) then Exit;

  HIDwriteDevice(DevList[CurrentDevice], Data); // send request to active device
  Sleep(10);

  for Start := 0 to SizeOF(Data) - 1 do Data[Start] := $00; // clear receive buffer
  Start := 0; // reset time out timer
  repeat
    OK := HIDreadDevice(DevList[CurrentDevice], Buffer);
    if not OK then
    begin
      Sleep(12); // <-----------
      INC(Start);
    end;
  until OK or // response received or
    (Start > 500); // time out   // <-----------

  if OK then // response received
    OK := (Buffer[1] = ACK) // target response received
  else
  begin
    INC(numerrors);
  end;
  Result := OK;
end;

procedure EraseBlocks(Address: Int64; NrOfBlocks, SizePerBlock: word); // block by block, one at a time.
var
  BlockNr: word;
begin
  BootloaderForm.ProgressBar.Max := NrOfBlocks;

  OutBuf[0] := STX;
  OutBuf[1] := cmdERASE;
  OutBuf[6] := 1; // one block a a time
  OutBuf[7] := 0; //

  for BlockNr := 0 to NrOfBlocks - 1 do
  begin

    OutBuf[2] := Address and $FF;
    OutBuf[3] := (Address shr 8) and $FF;
    OutBuf[4] := (Address shr 16) and $FF;
    OutBuf[5] := (Address shr 24) and $FF;

    SendDataWithAck(OutBuf, cmdErase);

    Address := Address + SizePerBlock;

    BootloaderForm.ProgressBar.Position := BlockNr;
    Application.ProcessMessages;
  end;
end;

procedure EraseDevice;
var
  Address,
    NrOfBlocks,
    Blocksize: Int64;
begin
  ShowInfo('Erasing flash memory.');

  case TargetDevice of

    mtPIC16, mtPIC18, mtPIC18FJ, mtPIC24, mtdsPIC:
      begin
        Blocksize := Bytes_to_Addresses(BootInfo.uiEraseBlock.intVal); // # addresses
        NrOfBlocks := BootInfo.ulBootStart.fValue div Blocksize;

        Address := 0;
        EraseBlocks(Address, NrOfBlocks, Blocksize);

        Sleep(100);
      end;

    mtPIC32:
      begin
        Blocksize := Bytes_to_Addresses(BootInfo.uiEraseBlock.intVal); // # addresses
        NrOfBlocks := (BootStartZeroBased div Blocksize);

        // Program section
        Address := $1D000000;
        EraseBlocks(Address, NrOfBlocks, BlockSize);
        Sleep(100);

        // Boot section
        ShowInfo('Erasing boot memory.');

        Address := $1FC00000;
        NrOfBlocks := (LowerBlockBoundary(EoBoot, BlockSize)) div Blocksize; // do not erase the option bytes
        EraseBlocks(Address, NrOfBlocks, BlockSize);

        Sleep(100);
      end;

  end;
  BootloaderForm.ProgressBar.Position := 0;
end;

procedure Send_WriteCmd_Address_BlockSize(Address: int64; BlockSize: word);
var
  HiAddress, LoAddress: int64;
  OutBuf: THIDBuffer;
begin
  LoAddress := Address and $FFFF;
  HiAddress := (Address shr 16) and $FFFF;

  OutBuf[0] := STX;
  OutBuf[1] := cmdWRITE;
  OutBuf[2] := LoAddress and $FF;
  OutBuf[3] := LoAddress shr 8;
  Outbuf[4] := HiAddress and $FF;
  Outbuf[5] := HiAddress shr 8;
  OutBuf[6] := BlockSize and $FF; // # bytes
  Outbuf[7] := BlockSize shr 8; // # bytes

  HIDwriteDevice(ActiveDevice, OutBuf);
end;

procedure WriteBytes(Index: Longword; BlockSize: word; NrOfBytes: Longword;
  Image: array of byte; AddressOffset: longword = 0);
  // Preconditions:
  // - Blocksize must be a multiple of BootInfo.uiWriteBlock.intVal
  // - Blocksize may not be larger than BootInfo.uiEraseBlock.intVal  <.. is this true?
  // - Index must be a bytenumber corresponding to an address at a ootInfo.uiWriteBlock.intVal boundary
var
  TotalBytesleft, BytesLeftInBlock: Longword;
  ByteNr, Tmp, BlockCnt: word;
  DataBuffer: THIDbuffer;
begin
  BlockCnt := 0;
  BootloaderForm.ProgressBar.Max := (NrOfBytes div BlockSize) + 1;

  NrOfBytes := UpperBlockBoundary(NrOfBytes, BootInfo.uiWriteBlock.intVal); // round up to boundary of uiWriteBlock size

  TotalBytesLeft := NrOfBytes;

  while TotalBytesLeft > 0 do
  begin

    // send block
    if TotalBytesLeft > BlockSize
      then BytesLeftInBlock := BlockSize
    else BytesLeftInBlock := TotalBytesLeft;

    Send_WriteCmd_Address_BlockSize(AddressOffset + Bytes_To_Addresses(Index), BytesLeftInBlock);

    // send HID buffer
    while BytesLeftInBlock > 0 do
    begin
      if BytesLeftInBlock > SizeOf(THIDBuffer)
        then Tmp := SizeOf(THidBuffer)
      else Tmp := BytesLeftInBlock;

      for ByteNr := 0 to SizeOf(THidBuffer) - 1 do DataBuffer[ByteNr] := $FF; // Clear HID buffer

      for ByteNr := 0 to Tmp - 1 do // copy data to HIDbuffer
      begin
        DataBuffer[ByteNr] := Image[Index];
        Inc(Index);
        Dec(BytesLeftInBlock);
        Dec(TotalBytesLeft);
      end;

      if BytesLeftInBlock > 0 // not the last HID block
        then SendDataNoAck(DataBuffer)
      else SendDataWithAck(DataBuffer, cmdWrite); // Last HID block
    end;

    inc(BlockCnt);
    BootloaderForm.ProgressBar.Position := BlockCnt;
    Application.ProcessMessages;

  end;
end;

procedure WriteFlashImage;
const
  MagicP32: array[0..3] of LongWord = ($3C1EBFC0, $37DE0050, $03C00008, $70000000);

var
  BlockSize, NrOfBlocks: word;
  Block: Integer;
  ImageAddress, Offset: LongWord;
  BS: int64;
  ResetVectorSize: byte;

  function DivUp(a, b: int64): int64;
  begin
    Result := a div b;
    if a mod b > 0 then inc(Result);
  end;

  function Between(Val_, Min_, Max_: int64): int64;
  begin
    Result := Val_;
    Result := max(Result, Min_); // Result never less than Min_
    Result := min(Result, Max_); // Result never greater than Max_
  end;

begin
  ShowInfo('Flashing program image.');
  BootloaderForm.ProgressBar.Max := EOFlash + 1;

  case TargetDevice of

    mtPIC16: // <---------------- only tested on a virtual P16 !!!!!!!!!!!!!!
      begin
        // prepare application reset vector

        ResetVectorSize := Reset_Vector_Size_P16;
        ImageAddress := Addresses_To_bytes(BootInfo.ulBootStart.fValue) - ResetVectorSize;

        FlashImage[ImageAddress] := FlashImage[0]; // Goto main program vector
        FlashImage[ImageAddress + 1] := FlashImage[1];

        BS := BootInfo.ulBootStart.fValue;

        FlashImage[0] := BS and $FF; //  Goto to bootloader vector
        FlashImage[1] := $2C or (BS shr 8);

        // Write flash data

        BlockSize := $100; // dummy big blocksize, handled by the Bootloader FW
        BlockSize := Between(BlockSize, BootInfo.uiWriteBlock.intVal, BootInfo.uiEraseBlock.intVal);
        BlockSize := UpperBlockBoundary(BlockSize, BootInfo.uiWriteBlock.intVal);

        // procedure WriteBytes(Index: Longword; BlockSize: word; NrOfBytes: Longword;
        //                      Image: array of byte; AddressOffset: longword = 0);

        WriteBytes(0, BlockSize, EoFlash + 1, FlashImage);

        // write reset vector area

        Blocksize := BootInfo.uiWriteBlock.intVal;
        NrOfBlocks := DivUp(ResetVectorSize, BlockSize);
        ImageAddress := Addresses_To_bytes(BootInfo.ulBootStart.fValue) - (BlockSize * NrOfBlocks); // # bytes

        // procedure WriteBytes(Index: Longword; BlockSize: word; NrOfBytes: Longword;
        //                      Image: array of byte; AddressOffset: longword = 0);

        WriteBytes(ImageAddress, BlockSize, NrOfBlocks * BootInfo.uiWriteBlock.intVal, FlashImage);

      end;

    mtPIC18, mtPIC18FJ:
      begin
        // prepare application reset vector

        ResetVectorSize := Reset_Vector_Size_P18;

        // copy the original jump to the main program (address 0) to just before the bootloader
        ImageAddress := Addresses_To_Bytes(BootInfo.ulBootStart.fValue) - ResetVectorSize;

        FlashImage[ImageAddress + 0] := FlashImage[0];
        FlashImage[ImageAddress + 1] := FlashImage[1];
        FlashImage[ImageAddress + 2] := FlashImage[2];
        FlashImage[ImageAddress + 3] := FlashImage[3];

        BS := BootInfo.ulBootStart.fValue div 2; // for calculated jumps: P18 addresses are in words !

        FlashImage[0] := BS and $FF;
        FlashImage[1] := $EF;
        FlashImage[2] := (BS shr 8) and $7F;
        FlashImage[3] := $F0 or ((BS shr 16) and $0F);

       // Write flash data

        BlockSize := $100; // dummy big blocksize, handled by the Bootloader FW
        BlockSize := Between(BlockSize, BootInfo.uiWriteBlock.intVal, BootInfo.uiEraseBlock.intVal);
        BlockSize := UpperBlockBoundary(BlockSize, BootInfo.uiWriteBlock.intVal);

       // procedure WriteBytes(Index: Longword; BlockSize: word; NrOfBytes: Longword;
       //                      Image: array of byte; AddressOffset: longword = 0);

        WriteBytes(0, BlockSize, EoFlash + 1, FlashImage);

        // write reset vector area

        Blocksize := BootInfo.uiWriteBlock.intVal;
        NrOfBlocks := DivUp(ResetVectorSize, BlockSize);
        ImageAddress := Addresses_To_bytes(BootInfo.ulBootStart.fValue) - (BlockSize * NrOfBlocks); // # bytes

        // procedure WriteBytes(Index: Longword; BlockSize: word; NrOfBytes: Longword;
        //                      Image: array of byte; AddressOffset: longword = 0);

        WriteBytes(ImageAddress, BlockSize, NrOfBlocks * BlockSize, FlashImage);

      end;

    mtPIC24, mtDsPIC:
      begin
        // prepare application reset vector

        ResetVectorSize := Reset_Vector_Size_P24;

        // copy the original jump to the main program (address 0) to just before the bootloader
        ImageAddress := Addresses_To_Bytes(BootInfo.ulBootStart.fValue) - ResetVectorSize;

        FlashImage[ImageAddress + 0] := FlashImage[0];
        FlashImage[ImageAddress + 1] := FlashImage[1];
        FlashImage[ImageAddress + 2] := FlashImage[2];
        FlashImage[ImageAddress + 3] := 0;
        FlashImage[ImageAddress + 4] := 0;
        FlashImage[ImageAddress + 5] := 0;

        // Replace the original jump with a jump to the bootloader
        BS := BootInfo.ulBootStart.fValue;

        FlashImage[0] := Lo(BS); // address
        FlashImage[1] := Hi(BS); // address
        FlashImage[2] := $04; //  Jump
        FlashImage[3] := 0;
        FlashImage[4] := 0;
        FlashImage[5] := 0;

        // write flash data

        BlockSize := $100; // dummy big blocksize, handled by the Bootloader FW
        BlockSize := Between(BlockSize, BootInfo.uiWriteBlock.intVal, BootInfo.uiEraseBlock.intVal);
        BlockSize := UpperBlockBoundary(BlockSize, BootInfo.uiWriteBlock.intVal);

        // procedure WriteBytes(Index: Longword; BlockSize: word; NrOfBytes: Longword;
        //                      Image: array of byte; AddressOffset: longword = 0);

        WriteBytes(0, BlockSize, EoFlash + 1, FlashImage);

        // write reset vector area

        Blocksize := BootInfo.uiWriteBlock.intVal;
        NrOfBlocks := DivUP(ResetVectorSize, BlockSize);
        ImageAddress := Addresses_To_bytes(BootInfo.ulBootStart.fValue) - (BlockSize * NrOfBlocks); // # bytes

        // procedure WriteBytes(Index: Longword; BlockSize: word; NrOfBytes: Longword;
        //                      Image: array of byte; AddressOffset: longword = 0);

        WriteBytes(ImageAddress, BlockSize, NrOfBlocks * BlockSize, FlashImage);

      end;

    mtPIC32:
      begin

        // prepare application reset vector

        ResetVectorSize := Reset_Vector_Size_P32;

        ImageAddress := (Addresses_To_bytes(BootStartZeroBased) - ResetVectorSize);
        for Block := 0 to 3 do
        begin

          FlashImage[ImageAddress + 0] := (MagicP32[Block]) and $FF;
          FlashImage[ImageAddress + 1] := (MagicP32[Block] shr 8) and $FF;
          FlashImage[ImageAddress + 2] := (MagicP32[Block] shr 16) and $FF;
          FlashImage[ImageAddress + 3] := (MagicP32[Block] shr 24) and $FF;

          inc(ImageAddress, 4); // point to next block

        end;

        // write flash data

        Offset := $1D000000; // offset between physical addresses and FlashImage index

        BlockSize := $100; // dummy big blocksize, handled by the Bootloader FW
        BlockSize := Between(BlockSize, BootInfo.uiWriteBlock.intVal, BootInfo.uiEraseBlock.intVal);
        BlockSize := UpperBlockBoundary(BlockSize, BootInfo.uiWriteBlock.intVal);

        // procedure WriteBytes(Index: Longword; BlockSize: word; NrOfBytes: Longword;
        //                      Image: array of byte; AddressOffset: longword = 0);

        WriteBytes(0, BlockSize, EoFlash + 1, FlashImage, Offset);

        // write reset vector area

        ShowInfo('Flashing boot vectors.');

        Blocksize := BootInfo.uiWriteBlock.intVal;
        NrOfBlocks := DivUp(ResetVectorSize, BlockSize);

        ImageAddress := Addresses_To_bytes(BootStartZeroBased) - (BlockSize * NrOfBlocks);

        // procedure WriteBytes(Index: Longword; BlockSize: word; NrOfBytes: Longword;
        //                      Image: array of byte; AddressOffset: longword = 0);

        WriteBytes(ImageAddress, BlockSize, NrOfBlocks * BlockSize, FlashImage, Offset);

      end;
  end; // case

  BootloaderForm.ProgressBar.Position := 0;
end;

procedure WriteBootImage; // Only for P32
type
  TData = record
    Address: Word;
    Data: LongWord;
  end;

const
  MagicP32Boot: array[0..3] of Tdata =
  (
    (Address: $0040; Data: $3C1E0000), // low order word of data not used any more (calculated now)
    (Address: $0044; Data: $37DE0000), // low order word of data not used any more (calculated now)
    (Address: $0048; Data: $03C00008),
    (Address: $004C; Data: $70000000)
    );

var
  Block: Integer;
  Offset: LongWord;
  BS: int64;
begin
  // Insert boot vectors into image from hex file

  BS := BootInfo.ulBootStart.Fvalue;

  for Block := 0 to 3 do

    with MagicP32Boot[Block] do
    begin

      if Block = 0 then // high order word of bootloader start
      begin
        BootImage[Address] := lo(BS shr 16);
        BootImage[Address + 1] := hi(BS shr 16);
      end
      else
        if Block = 1 then // low order word of bootloader start
      begin
        BootImage[Address] := lo(BS);
        BootImage[Address + 1] := hi(BS);
      end
      else
      begin // Fixed data
        BootImage[Address] := (Data) and $FF;
        BootImage[Address + 1] := (Data shr 8) and $FF;
      end;

       // fixed data
      BootImage[Address + 2] := (Data shr 16) and $FF;
      BootImage[Address + 3] := (Data shr 24) and $FF;

    end;

  ShowInfo('Flashing boot image.');

  Offset := $1FC00000;

  // procedure WriteBytes(Index: Longword; BlockSize: word; NrOfBytes: Longword;
  //                      Image: array of byte; AddressOffset: longword = 0);

  WriteBytes(0, $100, (EoBoot + 1) - BootInfo.uiEraseBlock.intVal, BootImage, Offset); // do not overwrite the optionbytes

  BootloaderForm.ProgressBar.Position := 0;
end;

function ParseFile(FName: string): Boolean;
var
  HEXfile: TextFile;
  FileRecord,
    RecordSegment: string;
  Cnt: LongInt;
  HEXcounter, P24counter: Integer;
  Offset: Int64;
  Address: Int64;
  NumBytes: Byte;
  RecType: Byte;
  Data: Byte;
  EOFile: Boolean;
  NotSupported: Int64;
  TooLarge: boolean;
  TmpFlashImage: TFlashImage;
  //TmpS: string;
  Reset_Vector_Size: word;

begin

  for Cnt := $0 to $7FFFF do FlashImage[Cnt] := $FF;
  for Cnt := $0 to $7FFFF do TmpFlashImage[Cnt] := $FF;
  for Cnt := $0 to $2FFF do BootImage[Cnt] := $FF;

  if TargetDevice = mtPIC16 then
    for Cnt := $0 to $7FFFF do
      if (Cnt and 1) = 1 then FlashImage[Cnt] := $3F;

  ShowInfo('Parsing HEX file: ' + ExtractFilename(FName));
  EOFlash := 0;
  EOBoot := 0;
  Offset := 0; // address offset from Extended Lineair Address record
  EOfile := FALSE; // end of file indicator from End Of File record
  Result := TRUE; // assume success
  NotSupported := 0;
  TooLarge := false;

  case TargetDevice of
    mtPic16: Reset_Vector_Size := Reset_Vector_Size_P16;
    mtPIC18, mtPIC18FJ: Reset_Vector_Size := Reset_Vector_Size_P18;
    mtPic24, mtDsPic: Reset_Vector_Size := Reset_Vector_Size_P24;
    mtPIC32: Reset_Vector_Size := Reset_Vector_Size_P32;
  else Reset_Vector_Size := 32;
  end;

  AssignFile(HEXfile, FName);
{$I-}
  Reset(HEXfile); // open file for reading

  BootloaderForm.ProgressBar.Max := FileSize(HEXfile);
  while not EOF(HEXfile) and // "real" end of file
    not EOFile and // end of file from Enf Of File record
    Result do // no errors found
  begin
    ReadLn(HEXfile, FileRecord); // read record from file
    //TmpS := FileRecord;
    BootloaderForm.ProgressBar.Position := FilePos(HEXfile);
    Application.ProcessMessages;

    if (FileRecord[1] = ':') then // valid record
    begin
      Delete(FileRecord, 1, 1);
      RecordSegment := Copy(FileRecord, 1, 2);
      NumBytes := 0;
      try
        NumBytes := StrToInt('$' + RecordSegment);
      except
        Result := FALSE;
      end;

      if Result and (NumBytes > 0) then // number of bytes properly read
      begin
        Delete(FileRecord, 1, 2);
        RecordSegment := Copy(FileRecord, 1, 4);
        Address := 0;
        try
          Address := StrToInt('$' + RecordSegment);
        except
          Result := FALSE;
        end;

        if Result then // address part properly read
        begin
          Delete(FileRecord, 1, 4);
          RecordSegment := Copy(FileRecord, 1, 2);
          RecType := 0;
          try
            RecType := StrToInt('$' + RecordSegment);
          except
            Result := FALSE;
          end;

          if Result then // Record type properly read
          begin
            Delete(FileRecord, 1, 2);
            case RecType of // process the different record types
              0: // Data record
                begin
                  for Cnt := 0 to NumBytes - 1 do
                  begin
                    RecordSegment := Copy(FileRecord, 1, 2);
                    Data := 0;
                    try
                      Data := StrToInt('$' + RecordSegment);
                    except
                      Result := FALSE;
                    end;
                    if Result then
                    begin
                      case TargetDevice of

                        mtPIC16:
                          begin
                          // important: the addresses in the hex file are the double of normal ones in the PIC16
                            if ((Offset shl 16) + Address + Cnt) div 2 < BootInfo.ulBootStart.fValue - Bytes_to_Addresses(Reset_Vector_Size) then
                            begin
                              FlashImage[(Offset shl 16) + Address + Cnt] := Data;
                              if ((Offset shl 16) + Address + Cnt) > EOflash then
                                EOFlash := (Offset shl 16) + Address + Cnt;
                            end
                            else
                            begin
                              if ((Offset shl 16) + Address + Cnt) div 2 < BootInfo.ulBootStart.fValue + $100 then
                                TooLarge := true;
                            end;
                          end;

                        mtPIC18, mtPIC18FJ:
                          begin
                            if ((Offset shl 16) + Address + Cnt) < BootInfo.ulBootStart.fValue - Bytes_to_Addresses(Reset_Vector_Size) then
                            begin
                              FlashImage[(Offset shl 16) + Address + Cnt] := Data;
                              if ((Offset shl 16) + Address + Cnt) > EOflash then
                                EOFlash := (Offset shl 16) + Address + Cnt;
                            end
                            else
                            begin
                              if ((Offset shl 16) + Address + Cnt) < BootInfo.ulBootStart.fValue + $100 then
                                TooLarge := true;
                            end;
                          end;

                        mtPIC24, mtdsPIC:
                          begin
                            //important: the addresses in the hex file are the double of normal ones in the PIC24
                            if (((Offset shl 16) + Address + Cnt) div 2) < BootInfo.ulBootStart.fValue - Bytes_to_Addresses(Reset_Vector_Size) then
                            begin
                              FlashImage[(Offset shl 16) + Address + Cnt] := Data;
                              if (((Offset shl 16) + Address + Cnt) > EOFlash) then
                                EOFlash := (Offset shl 16) + Address + Cnt;
                            end
                            else
                            begin
                              if (((Offset shl 16) + Address + Cnt) div 2) < BootInfo.ulBootStart.fValue + $100 then
                                TooLarge := true;
                            end;
                          end;

                        mtPIC32:
                          begin
                            if (Offset >= $1FC0) then // data for boot section, max 64K assumed
                            begin
                              BootImage[Address + Cnt] := Data; // no range check possible here (size of boot flash not known)
                              if ((Address + Cnt) > EOBoot) then EOBoot := Address + Cnt;
                            end
                            else // data for program image
                            begin
                              if (((Offset and $00FF) shl 16) + Address + Cnt) <
                                BootStartZeroBased - Bytes_to_Addresses(Reset_Vector_Size) then
                              begin
                                FlashImage[((Offset and $00FF) shl 16) + Address + Cnt] := Data;
                                if (((Offset and $00FF) shl 16) + Address + Cnt) > EOFlash
                                  then EOFlash := ((Offset and $00FF) shl 16) + Address + Cnt;
                              end
                              else
                              begin
                                if (((Offset and $00FF) shl 16) + Address + Cnt) <
                                  BootStartZeroBased + $100 then
                                  TooLarge := true;
                              end;
                            end;
                          end;

                      end; // case

                      Delete(FileRecord, 1, 2);
                    end
                    else Break;
                  end;
                end;
              1: // end Of Fle record
                begin
                  EOFile := TRUE;
                end;
              2: // Extended Segment Address record
                begin
                  INC(NotSupported);
                end;
              3: // Start Segment Address record
                begin
                  INC(NotSupported);
                end;
              4: // Extended Linear Address record
                begin
                  RecordSegment := Copy(FileRecord, 1, 4);
                  try
                    Offset := StrToInt('$' + RecordSegment);
                  except
                    Result := FALSE;
                  end;
                end;
              5: // Start Lineair Adress record
                begin
                  INC(NotSupported);
                end;
            end;
          end; // record type
        end; // address
      end; // number of bytes
    end; // valid record
  end; // eof file

  CloseFile(HEXfile);

  // Convert image into P24 format if necessary

  if (TargetDevice = mtPIC24) or (TargetDevice = mtdsPIC) then
  begin

    HEXcounter := 0;
    P24counter := 0;

    repeat
      TmpFlashImage[P24counter + 0] := FlashImage[HEXcounter + 0]; //  These 3 bytes go to the bootvector
      TmpFlashImage[P24counter + 1] := FlashImage[HEXcounter + 1]; //  and must be replaced by the original
      TmpFlashImage[P24counter + 2] := FlashImage[HEXcounter + 2]; //  bootloader bytes

      INC(P24counter, 3);
      INC(HEXcounter, 4); // do leave out every fourth byte (dummy byte)
    until (HEXcounter > EOFlash);

    // adjust flash size
    EOFlash := P24counter;

    for Cnt := $0 to $7FFFF do FlashImage[Cnt] := $FF;
    for Cnt := 0 to EoFlash do FlashImage[Cnt] := TmpFlashImage[cnt]; // copy back to original array

  end;

  // EoFlash and EoBoot are a byte numbers in FlashImage and BootImage (can differ from the address in the code)

  if TooLarge or (EOFlash > (Addresses_To_Bytes(BootInfo.ulBootStart.fValue and $00FFFFFF)) - Reset_Vector_Size) then // the "and" is for P32 addresses
  begin
    ShowInfo('  Image is too big for this PIC and');
    ShowInfo('  would overwrite (part of) bootloader!');
    ShowInfo('  Programming will be aborted.');
    Result := FALSE;
  end;

  if not Result then ShowInfo('Error parsing hex file.');
  if (NotSupported <> 0) then ShowInfo(IntToStr(NotSupported) + ' unsupported entries found!');

  BootloaderForm.ProgressBar.Position := 0;

  if Result then
    if TargetDevice <> mtPIC32
      then ShowInfo('   EoFlash: Address 0x' + intToHex(Bytes_To_Addresses(EOFlash), 6))
    else
    begin
      ShowInfo('   EoFlash: Address 0x' + intToHex(EOFlash or $9D000000, 6));
      ShowInfo('   EoBoot : Address 0x' + intToHex(EOBoot or $9FC00000, 6));
    end;

end;

procedure testboot;
begin
  checkloader;
  SendCommandNoAck(cmdReboot, OutBuf);
end;

var // free user blocks values
  NrOfUserBlocks, NrOfUserAddresses, NrOfUserBytes: int64;
  FirstBlockAboveProgramendAddr, OneBlockbelowBootloaderStartAddr, TwoBlocksBelowBootloaderStartAddr: int64;
  UserFlashStartAddress, UserFlashEndAddress, BootFlagAddress: int64;
  UserBlocks: array of int64;

procedure ShowUserBlocks;
var
  I: word;
begin
  if NrOfUserBlocks <= 0 then
  begin
    ShowInfo('NO Free User Flash');
    ShowInfo('NO Bootflag place');
    exit;
  end;

  ShowInfo('Free User Flash size         : ' + inttostr(NrOfUserBytes) + ' bytes');
  ShowInfo('Free User Flash Start Address: 0x' + inttohex(UserFlashStartAddress, 8));
  ShowInfo('Free User Flash End Address  : 0x' + inttohex(UserFlashEndAddress, 8));
  ShowInfo('BootFlag Address             : 0x' + inttohex(BootFlagAddress, 8));

  ShowInfo('Free User Flash (' + inttostr(NrOfUserBlocks) + ' blocks of ' + inttostr(BootInfo.uiEraseBlock.intVal) + ' bytes):');
  if NrOfUserBlocks <= 5 then
  begin
    for I := 0 to (NrOfUserBlocks - 1) do ShowInfo('   0x' + intTohex(UserBlocks[I], 8));
  end else
  begin
    for I := 0 to 1 do ShowInfo('   0x' + intTohex(UserBlocks[I], 8));
    ShowInfo('   ...');
    for I := NrOfUserBlocks - 2 to NrOfUserBlocks - 1 do ShowInfo('   0x' + intTohex(UserBlocks[I], 8));
  end;
end;

procedure CalculateUserBlocks;
var
  BlockSizeInAddresses, Cnt: int64;
  I: word;
begin
  BlockSizeInAddresses := Bytes_to_Addresses(BootInfo.uiEraseBlock.intVal);

  // first block that can be used by the user
  FirstBlockAboveProgramEndAddr := UpperBlockBoundary(Bytes_To_Addresses(EoFlash) + 1, BlockSizeInAddresses);
  if TargetDevice = mtPIC32 then FirstBlockAboveProgramEndAddr := FirstBlockAboveProgramEndAddr or $9D000000;

  OneBlockBelowBootloaderStartAddr := LowerBlockBoundary((BootInfo.ulBootStart.fValue - 1) and $00FFFFFF, BlockSizeInAddresses);
  if TargetDevice = mtPIC32 then OneBlockBelowBootloaderStartAddr := OneBlockBelowBootloaderStartAddr or $9D000000;

  // last block that can be used by the user
  TwoBlocksBelowBootloaderStartAddr := OneBlockBelowBootloaderStartAddr - BlockSizeInAddresses;
  if TargetDevice = mtPIC32 then TwoBlocksBelowBootloaderStartAddr := TwoBlocksBelowBootloaderStartAddr or $9D000000;

  NrOfUserAddresses := (OneBlockBelowBootloaderStartAddr - FirstBlockAboveProgramEndAddr); // the 2nd block below the bootloader can be used
  NrOfUserBlocks := NrOfUserAddresses div BlockSizeInAddresses;

  if NrOfUserBlocks <= 0 then exit;

  NrOfUserBytes := Addresses_to_Bytes(NrOfUserAddresses);

  UserFlashStartAddress := FirstBlockAboveProgramEndAddr;
  UserFlashEndAddress := OneBlockBelowBootloaderStartAddr - 1;

  BootFlagAddress := TwoBlocksBelowBootloaderStartAddr;

  Setlength(UserBlocks, 0);
  Cnt := UserFlashStartAddress;
  for I := 0 to NrOfUserBlocks - 1 do
  begin
    SetLength(UserBlocks, I + 1);
    UserBlocks[I] := Cnt;
    inc(Cnt, BlockSizeInAddresses);
  end;

end;

procedure LoadApplication(FName: string);
var
  Result: boolean;
begin
  State := 100;

  Result := ParseFile(FName);

  if Result then
  begin
    EraseDevice;

    WriteFlashImage;
    if TargetDevice = mtPIC32 then WriteBootImage;

    ShowInfo('Flashing complete.');

    CalculateUserBlocks;
    ShowUserBlocks;

    if Setup.DisableBL then
    begin
      SendCommandNoAck(cmdSetBootFlag, OutBuf);
      ShowInfo('Disabling boot loader...');
    end
    else SendCommandNoAck(cmdReboot, OutBuf);
  end;
  HIDcloseDevice(Target);

  ClearRestore;
  State := 0;
  LoaderActive := FALSE;
  BootLoaderActivated := FALSE;
  DeviceConnected := False;

  ShowInfo('Completed.');

  {
  SetRestore;
  ShowInfo('Reboot......');
  InStr := '';
  DeviceConnected := FALSE;
  TimStart := -1;
  BootLoaderActivated := FALSE;}

  {
  USBsetEventHandler(@CheckLoader); // assign callback procedure for USB change event
  State := 0;
  FindDevices;
  }
end;

function aLen(S: string; Len: integer): string;
begin
  if (Length(S) > Len) then S := Copy(S, 1, 20)
  else
    if (Length(S) = Len) then
  else
    repeat
      S := S + ' ';
    until (Length(S) >= Len);
  Result := S;
end;

procedure CheckLoader; stdcall;
// Callback procedure for USB change event
var
  Found: Boolean;
  NumDevices: Byte;
  IndexA, IndexB: byte;
  FoundEntry: boolean;
  Str: string;
begin
  if BootloaderActivated then Exit; // bootloader mode is already active

  BootloaderActivated := FALSE;
  CurrentDevice := -1; // reset variable

  HIDscanForDevices(DevList,
    NumDevices,
    Setup.VID,
    Setup.PID); // try to locate target device
  if (NumDevices = 1) then CurrentDevice := 0;

  if (CurrentDevice = 0) then // our target device (VID/PID) is found
  begin
    ShowInfo('HID Target device found');
    //HIDsetEventHandler(nil); // remove any reader callback
    if HIDopenDevice(DevList[CurrentDevice]) then // try to open target device
    begin
      ShowInfo('HID device successfully opened');
      Sleep(200);

      SendCommand_GetReply(cmdINFO, OutBuf); // send cmINFO command

      LoadBootRecord(OutBuf, BootInfo); // load boot record from received data
      Found := (BootInfo.bMcuType.fFieldType = bifMCUTYPE) and // MCU type field ID
      (BootInfo.bMcuType.fValue in [mtPIC16, mtPIC18, mtPIC18FJ, mtPIC24, mtDSPIC, mtPIC32]); // correct MCU type

      if Found then // target MCU type is found
      begin
        ShowInfo('Target MCU found.');

        ShowBootInfo(BootInfo); // show received bootloader info
        SendCommand_GetReply(cmdBOOT, OutBuf); //send cmBOOT command
        BootloaderActivated := (OutBuf[1] = cmdBOOT); // ACK received on boot command

        if BootloaderActivated then // target is ready to be flashed
        begin
          USBsetEventHandler(nil); // clear callback procedure for USB change event
          ShowInfo('Target device in boot load mode.');
          ShowInfo('   ');
          ShowInfo('Press FLASH button to start flashing.');
          ShowInfo('   ');

          Bootloaderform.GetFileBut.Enabled := true;
          Bootloaderform.HistoryBut.Enabled := true;

          case TargetDevice of

            mtPIC16:
              begin
                if not FileExists(Setup.Hexfilename_P16) then BootloaderForm.GetFileButClick(nil);
                if not FileExists(Setup.Hexfilename_P16) then BootloaderForm.OD.Filename := Setup.Hexfilename_P16;
                BootloaderForm.SL.Caption := Setup.Hexfilename_P16;
              end;

            mtPIC18, mtPIC18FJ:
              begin
                if not FileExists(Setup.Hexfilename_P18) then BootloaderForm.GetFileButClick(nil);
                if not FileExists(Setup.Hexfilename_P18) then BootloaderForm.OD.Filename := Setup.Hexfilename_P18;
                BootloaderForm.SL.Caption := Setup.Hexfilename_P18;
              end;

            mtPIC24, mtdsPIC:
              begin
                if not FileExists(Setup.Hexfilename_P24) then BootloaderForm.GetFileButClick(nil);
                if not FileExists(Setup.Hexfilename_P24) then BootloaderForm.OD.Filename := Setup.Hexfilename_P24;
                BootloaderForm.SL.Caption := Setup.Hexfilename_P24;
              end;

            mtPIC32:
              begin
                if not FileExists(Setup.Hexfilename_P32) then BootloaderForm.GetFileButClick(nil);
                if not FileExists(Setup.Hexfilename_P32) then BootloaderForm.OD.Filename := Setup.Hexfilename_P32;
                BootloaderForm.SL.Caption := Setup.Hexfilename_P32;
              end;
          else
            begin
            end;
          end;

          if FileExists(BootloaderForm.Sl.Caption) then
            BootloaderForm.Od.InitialDir := ExtractFileDir(BootloaderForm.Sl.Caption);

        end
        else ShowInfo('Unable to enter boot load mode !');
      end
      else ShowInfo('Target MCU not found !');
    end
    else ShowInfo('Unable to open HID device !');
  end
  else
  begin
    //ShowInfo(' USB event......');

    Str := '';

    HIDscanForDevices(NewDevices, NewNumDevices, 0, 0);

    if (NewNumDevices > 0) and (OldNumDevices > 0) then
    begin // compare new usb list and old usb list

      for IndexA := 0 to NewNumDevices - 1 do // check if all devices in the new list are also present in the old one
      begin
        FoundEntry := false;
        for IndexB := 0 to OldNumDevices - 1 do
        begin
          if (NewDevices[IndexA].VID = OldDevices[IndexB].Vid) and
             (NewDevices[IndexA].PID = OldDevices[IndexB].Pid)
             then
             begin
               FoundEntry := true;
               break;
             end;
        end;
        if not FoundEntry then
        Str := Str +  (' (' + inttohex(NewDevices[IndexA].VID, 4) + ', ' + inttohex(NewDevices[IndexA].PID, 4) + ')');
      end;

      (*
      for IndexB := 0 to OldNumDevices - 1 do // check if all devices in the old list are also present in the new one
      begin
        FoundEntry := false;
        for IndexA := 0 to NewNumDevices - 1 do
        begin
          if (NewDevices[IndexA].VID = OldDevices[IndexB].Vid) and
             (NewDevices[IndexA].PID = OldDevices[IndexB].Pid)
             then
             begin
               FoundEntry := true;
               break;
             end;
        end;
        if not FoundEntry
        Str := Str +  (' (' + inttohex(OldDevices[IndexB].VID, 4) + ', ' + inttohex(OldDevices[IndexB].PID, 4) + ')');
      end;
      *)
    end;

    if Str > '' then
    begin
     Str := ' USB event... ' + Str;
     ShowInfo(Str);
    end;
     
     // copy the new list into the old one
    for IndexA := 0 to NewNumDevices - 1 do
      begin
        OldDevices[IndexA] := NewDevices[IndexA];
        OldNumDevices := NewNumDevices;
      end;
  
  end;
end;

initialization
finalization
  HIDsetEventHandler(nil); // remove any reader callback
  USBsetEventHandler(nil);
end.
