unit BL_Main;

(*
{$O-} // Optimisation OFF
{$D+} // Debug information ON
{$I+} // I/O checking ON
{$L+} // Local Symbols ON
{$Q+} // Overflow Checking ON
{$R+} // Range Checking ON
*)

{$IFDEF NoDebug}

{$O+} // Optimisation ON
{$D-} // Debug information OFF
{$I-} // I/O checking OFF
{$L-} // Local Symbols OFF
{$Q-} // Overflow Checking OFF
{$R-} // Range Checking OFF

{$ELSE}
{$O-} // Optimisation OFF
{$D+} // Debug information ON
{$I+} // I/O checking ON
{$L+} // Local Symbols ON
{$Q+} // Overflow Checking ON
{$R+} // Range Checking ON

{$ENDIF}
                                                       
interface

uses
  Windows,
  Messages,
  SysUtils,
  Variants,
  Classes,
  Graphics,
  Controls,
  Forms,
  Dialogs,
  StdCtrls,
  ExtCtrls,
  ComCtrls,
  HIDinterface,
  crd_Time,
  Crd_Edit,
  ButtonWithColor,
  Buttons,
  ScrollLabel;

type
  TBootloaderForm = class(TForm)
    InputWindow: TMemo;
    ConnectIndicator: TStaticText;
    UpdateTimer: TTimer;
    Panel2: TPanel;
    ProgressBar: TProgressBar;
    GetFileBut: TButton;                                                    
    Panel1: TPanel;
    SL: TScrollLabel;
    OD: TOpenDialog;
    GroupBox1: TGroupBox;
    StaticText1: TStaticText;
    StaticText2: TStaticText;
    VIDlabel: TStaticText;
    PIDlabel: TStaticText;
    Label1: TLabel;
    CheckBox2: TCheckBox;
    FlashBut: TBitBtnWC;
    ResetButton: TButton;
    HistoryBut: TButton;
    ListBox1: TListBox;
    procedure UpdateTimerEvent(Sender: TObject);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure FormActivate(Sender: TObject);
    procedure GetFileButClick(Sender: TObject);
    procedure VIDlabelMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure PIDlabelMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure CheckBox2Click(Sender: TObject);
    procedure FlashButClick(Sender: TObject);
    procedure ResetButtonClick(Sender: TObject);
    procedure HistoryButClick(Sender: TObject);
    procedure ListBox1DblClick(Sender: TObject);
    Procedure AddToHistory(Name: string);
    procedure ListBox1KeyDown(Sender: TObject; var Key: Word;
      Shift: TShiftState);
    procedure FormResize(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);

  private
    { Private declarations }
  public
    { Public declarations }
    procedure NewInit;
    procedure SaveSettings;
    procedure LoadSettings;

  end;

var
  BootloaderForm: TBootloaderForm;
  NumErrors: integer = 0;
  LoaderActive: Boolean = FALSE;

procedure ShowInfo(Msg: string);
procedure ClearWindow;
procedure SetRestore;
procedure ClearRestore;

implementation
uses
  BL_Vars,
  BL_Flash,
  MG_IO,
  RCS_Resize,
  SimpleIniFile;

{$R *.dfm}

var
  FirstInit: Boolean = TRUE;
  pMsg: string = '';
  InStr: string = '';
  MyResize: TMyResize;

procedure SetRestore;
begin
  USBsetEventHandler(@FindDevices); // assign callback to detect USB activity
end;

procedure ClearRestore;
begin
  USBsetEventHandler(nil); // assign callback to detect USB activity
  //HIDsetEventHandler(nil); // assign callback procedure for HID reader
end;

function StrToReal(Data: string): Double;
var
  N: Double;
  E: Integer;
begin
  VAL(Data, N, E);
  if (E = 0)
    then StrToReal := N
  else StrToReal := 0;
end;

procedure ShowInfo(Msg: string);
const
  Len = 47;
var
  N: Integer;
  S: string;
begin
  if (Msg = pMsg) then Exit;
  if (Msg[1] = ' ') and
    (Msg <> pMsg) then
  begin
    pMsg := Msg;
    Delete(Msg, 1, 1);
    BootLoaderForm.InputWindow.Lines.Add(Msg);
    Exit;
  end;
  if pMsg <> Msg then
  begin
    pMsg := Msg;
    if (Length(Msg) > Len) then
    begin
      Msg := Copy(Msg, 1, 44) + '...';
    end;
    N := TimerMS - TimStart;
    S := IntToStr(N);

    if (Length(S) < 6) then
      repeat
        S := ' ' + S;
      until (Length(S) >= 6);

    if (Length(Msg) < Len) then
      repeat
        Msg := Msg + ' ';
      until (Length(Msg) >= Len);
    if (TimStart = -1) then BootLoaderForm.InputWindow.Lines.Add(Msg)
    else BootLoaderForm.InputWindow.Lines.Add(Msg + '   [' + S + ' ms ]');
  end;
end;

procedure InitialText;
begin
  BootLoaderForm.InputWindow.Lines.Clear;
  ShowInfo('                 Press PIC RESET button to get into boot load mode.');
  ShowInfo('   ');
end;

procedure ResetText;
begin
  BootLoaderForm.InputWindow.Lines.Clear;
  InitialText;
end;

procedure CheckText;
begin
  BootLoaderForm.InputWindow.Lines.Clear;
  ShowInfo('   ');
  ShowInfo('   ');
  ShowInfo('   ');
  ShowInfo('               Press Check button to show all HID device ID''s')
end;

procedure ClearWindow;
begin
  BootLoaderForm.InputWindow.Lines.Clear;
end;

procedure TBootloaderForm.UpdateTimerEvent(Sender: TObject);
begin

  if BootloaderActivated and (State = 100) then
  begin
    ConnectIndicator.Color := clLime;
    ConnectIndicator.Caption := 'Flashing';
    ResetButton.Enabled := FALSE;
    FlashBut.Enabled := FALSE;
  end
  else
    if BootloaderActivated then
  begin
    ConnectIndicator.Color := clOrange;
    ConnectIndicator.Caption := ' Target connected ';
    ResetButton.Enabled := TRUE;

    if FileExists(SL.Caption)
      then
    begin
      FlashBut.Enabled := TRUE;
      SL.Font.Color := clWhite;
    end
    else
    begin
      FlashBut.Enabled := FALSE;
      SL.Font.Color := clRed;
      if trim(Sl.Caption) = '' then SL.Caption := 'HexFile?'
    end;

  end
  else
  begin
    ConnectIndicator.Color := clRed;
    ConnectIndicator.Caption := '------';
    ResetButton.Enabled := TRUE;
    FlashBut.Enabled := FALSE;
  end;
  
  Application.ProcessMessages;
end;

procedure TBootloaderForm.SaveSettings;
var
  F: file of TSetup;
begin
  with Setup do
  begin
    Xb := Left;
    Yb := Top;
  end;
  AssignFile(F, ProgDir + SettingsFilename);
{$I-}
  Rewrite(F);
  Write(F, Setup);
  CloseFile(F);
end;

procedure TBootloaderForm.NewInit;
begin
  with Setup do
  begin
    Xb := (Screen.Width div 2) - (Width div 2);
    Yb := (Screen.Height div 2) - (Height div 2);
    VID := $3930;
    PID := $2021;
    HEXfilename_P16 := ProgDir + '';
    HEXfilename_P18 := ProgDir + '';
    HEXfilename_P24 := ProgDir + '';
    HEXfilename_P32 := ProgDir + '';
    DisableBL := FALSE;
  end;
end;

procedure TBootloaderForm.LoadSettings;
var
  F: file of TSetup;
begin
  AssignFile(F, ProgDir + SettingsFilename);
{$I-}
  Reset(F);
  if (IORESULT = 0) then
  begin
    Read(F, Setup);
    if (IORESULT = 0) then
    else NewInit;
    CloseFile(F);
  end
  else NewInit;
end;

function GetVersionStr(Fname: string; IncludeBuild: Boolean): string;
var
  InfoSize: DWord;
  Buffer: Pointer;
  FI: PVSFixedFileInfo;
  Dummy: Cardinal;
begin
  InfoSize := GetFileVersionInfoSize(PChar(Fname), Dummy);
  if (InfoSize <> 0) then
  begin
    GetMem(Buffer, InfoSize);
    try
      if IncludeBuild then
        if GetFileVersionInfo(PChar(Fname), 0, InfoSize, Buffer) then
          if VerQueryValue(Buffer, '\', Pointer(FI), InfoSize) then
            GetVersionStr := IntToStr(HiWord(FI^.dwFileVersionMS)) + '.' +
              IntToStr(LoWord(FI^.dwFileVersionMS)) + '.' +
              IntToStr(HiWord(FI^.dwFileVersionLS)) + '.' +
              IntToStr(LoWord(FI^.dwFileVersionLS))
          else
        else
      else
        if GetFileVersionInfo(PChar(Fname), 0, InfoSize, Buffer) then
        if VerQueryValue(Buffer, '\', Pointer(FI), InfoSize) then
          GetVersionStr := IntToStr(HiWord(FI^.dwFileVersionMS)) + '.' +
            IntToStr(LoWord(FI^.dwFileVersionMS)) + '.' +
            IntToStr(HiWord(FI^.dwFileVersionLS));
    finally
      FreeMem(Buffer, InfoSize);
    end;
  end;
end;

procedure TBootloaderForm.FormActivate(Sender: TObject);
var
  VersionStr: string;
begin

  ClearRestore;

  if FirstInit then
  begin
    VersionStr := GetVersionStr(Application.ExeName, FALSE);
    Caption := ' Universal PIC Flash utility:  V' + VersionStr;
    LoadSettings;
    OD.InitialDir := ProgDir;
    OD.Title := 'Load Flash image file';
    OD.Filter := 'Image files (.hex)|*.hex';
    OD.DefaultExt := '.hex';
      {
      If NOT FileExists(Setup.Hexfilename) then GetFileButClick(Nil);
      If NOT FileExists(Setup.Hexfilename) then OD.Filename   := Setup.Hexfilename;
      SL.Caption := Setup.Hexfilename;
      }
    BootloaderForm.SL.Caption := ' ';
    with Setup do
    begin
      Left := Xb;
      Top := Yb;
    end;
    VIDlabel.Caption := IntToHex(Setup.VID, 4);
    PIDlabel.Caption := IntToHex(Setup.PID, 4);
    InitialText;
    FirstInit := FALSE;
    CheckBox2.Checked := Setup.DisableBL;
    //fb.SetFocus;
  end;

  InitVars;

  if not TargetConnected then
  begin
    BootloaderActivated := FALSE;
    CurrentDevice := -1;
    LoaderActive := TRUE;
    ResetText;
  end

  else InitialText;

  USBsetEventHandler(@CheckLoader); // assign callback procedure for USB change event
end;

procedure TBootloaderForm.FormClose(Sender: TObject;
  var Action: TCloseAction);
begin
  SetRestore;
  SaveSettings;
  LoaderActive := FALSE;
end;

procedure TBootloaderForm.AddToHistory(Name: string);
var
  Fn: string;
  I: integer;
begin
  
  case TargetDevice of
    mtPIC16: Fn := 'History_P16.txt';
    mtPIC18, mtPIC18FJ: Fn := 'History_P18.txt';
    mtPIC24, mtDsPIC: Fn := 'History_P24.txt';
    mtPIC32: Fn := 'History_P32.txt';
  end;
  Fn := ProgDir + Fn;

  ListBox1.Items.Clear;

  if FileExists(Fn) then
  ListBox1.Items.LoadFromFile(Fn);

  I := ListBox1.Items.IndexOf(Name);
  if I < 0
    then ListBox1.Items.Insert(0, Name) // name not present yet, add it
  else ListBox1.Items.Exchange(0, I); // name already present, rearrange list

  if Listbox1.Items.Count > 10 then ListBox1.Items.Delete(9);
  ListBox1.Items.SaveToFile(Fn);

end;

procedure TBootloaderForm.GetFileButClick(Sender: TObject);
begin
  if OD.Execute then
    case TargetDevice of
      mtPIC16:
        begin
          Setup.Hexfilename_P16 := OD.FileName;
          SL.Caption := Setup.Hexfilename_P16;
        end;

      mtPic18, mtPIC18FJ:
        begin
          Setup.Hexfilename_P18 := OD.FileName;
          SL.Caption := Setup.Hexfilename_P18;
        end;

      mtPIC24, mtdsPIC:
        begin
          Setup.Hexfilename_P24 := OD.FileName;
          SL.Caption := Setup.Hexfilename_P24;
        end;

      mtPIC32:
        begin
          Setup.Hexfilename_P32 := OD.FileName;
          SL.Caption := Setup.Hexfilename_P32;
        end;
    end;

  AddToHistory(Sl.Caption);
  if FileExists(OD.FileName) then
  Od.InitialDir := ExtractFileDir(Od.FileName);

  //fb.SetFocus;
end;

procedure TBootloaderForm.VIDlabelMouseDown(Sender: TObject;
  Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  Setup.VID := StrToInt('$' + EditHexLong(Sender, TRUE, 0, $FFFF));
  VIDlabel.Caption := IntToHex(Setup.VID, 4);
end;

procedure TBootloaderForm.PIDlabelMouseDown(Sender: TObject;
  Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  Setup.PID := StrToInt('$' + EditHexLong(Sender, TRUE, 0, $FFFF));
  PIDlabel.Caption := IntToHex(Setup.PID, 4);
end;

procedure TBootloaderForm.CheckBox2Click(Sender: TObject);
begin
  with (Sender as TCheckBox) do Setup.DisableBL := Checked;
  USBsetEventHandler(@CheckLoader); // assign callback procedure for USB change event
  //fb.SetFocus;
end;

procedure TBootloaderForm.FlashButClick(Sender: TObject);
begin
  //fb.SetFocus;
  if not BootloaderActivated then
  begin
    ShowMessage('Bootloader is not active.');
    Exit;
  end;
  
  GetFileBut.Enabled := false;
  Historybut.Enabled := false;
  Application.ProcessMessages;

  case TargetDevice of
    mtPIC16: LoadApplication(Setup.HEXfilename_P16); // load application firmware
    mtPIC18, mtPIC18FJ: LoadApplication(Setup.HEXfilename_P18); // load application firmware
    mtPIC24, mtDsPIC: LoadApplication(Setup.HEXfilename_P24); // load application firmware
    mtPIC32: LoadApplication(Setup.HEXfilename_P32); // load application firmware
  end;

end;

procedure CtrlBootloader(Mode: Byte);
// Mode 0 : disable bootloader
//      1 : enable bootloader
//      2 : enablle bootloader and reset
var
  Cmd: string;
begin
//  If NOT Connected then Exit;
  Cmd := '         ';
  Cmd[1] := CHR($09); // length
  Cmd[2] := CHR($00);
  Cmd[3] := CHR($01); // user id
  Cmd[4] := CHR($01); // group id: system
  Cmd[5] := CHR($06); // upgrade
  Cmd[6] := CHR($03); // set mode
  Cmd[7] := CHR($01); // message number
  Cmd[8] := CHR(Mode); // set mode:
                        //  $00 - edisble bootloader
                        //  $01 - enable bootloader
                        //  $03 - enable bootloader and reset
  Cmd[9] := CHR($00); // crc byte
  SendData(Cmd);
end;

procedure TBootloaderForm.ResetButtonClick(Sender: TObject);
begin
(*
  //If FirstInit then Exit;
  FindDevices;
  Sleep(200);
  CtrlBootloader($03);
*)

  ClearRestore;
  InitVars;

  if not TargetConnected then
  begin
    BootloaderActivated := FALSE;
    CurrentDevice := -1;
    LoaderActive := TRUE;
    ResetText;
  end

  else InitialText;

  SL.Caption := '';

  Listbox1.Visible := False;

  USBsetEventHandler(@CheckLoader); // assign callback procedure for USB change event
end;

procedure TBootloaderForm.HistoryButClick(Sender: TObject);
var
  Fn: string;
begin

  case TargetDevice of
    mtPIC16: Fn := 'History_P16.txt';
    mtPIC18, mtPIC18FJ: Fn := 'History_P18.txt';
    mtPIC24, mtDsPIC: Fn := 'History_P24.txt';
    mtPIC32: Fn := 'History_P32.txt';
  end;
  Fn := ProgDir + Fn;

  ListBox1.Items.Clear;
  if FileExists(Fn) then
  begin
    ListBox1.Items.LoadFromFile(Fn);
    ListBox1.Visible := true;
  end;  

end;

procedure TBootloaderForm.ListBox1DblClick(Sender: TObject);
var
  Fn: string;
  I: integer;
begin
  I := ListBox1.ItemIndex;

  Fn := ListBox1.Items[I];
  Fn := Fn;

  if FileExists(Fn) then // do nothing if the file does not exist
  begin
    ListBox1.Items.Exchange(0, I);

    if ListBox1.Items.Count > 10 then ListBox1.Items.Delete(9);

    case TargetDevice of
      mtPIC16: ListBox1.Items.SaveToFile(ProgDir + 'History_P16.txt');
      mtPIC18, mtPIC18FJ: ListBox1.Items.SaveToFile(ProgDir + 'History_P18.txt');
      mtPIC24, mtDsPIC: ListBox1.Items.SaveToFile(ProgDir + 'History_P24.txt');
      mtPIC32: ListBox1.Items.SaveToFile(ProgDir + 'History_P32.txt');
    end;

    case TargetDevice of
      mtPIC16:
        begin
          Setup.Hexfilename_P16 := Fn;
          SL.Caption := Setup.Hexfilename_P16;
        end;

      mtPic18, mtPIC18FJ:
        begin
          Setup.Hexfilename_P18 := Fn;
          SL.Caption := Setup.Hexfilename_P18;
        end;

      mtPIC24, mtdsPIC:
        begin
          Setup.Hexfilename_P24 := Fn;
          SL.Caption := Setup.Hexfilename_P24;
        end;

      mtPIC32:
        begin
          Setup.Hexfilename_P32 := Fn;
          SL.Caption := Setup.Hexfilename_P32;
        end;
    end;
    
    Od.InitialDir := ExtractFileDir(Fn);

  end else ShowMessage('File '+ Fn + ' does not exist');

  ListBox1.Visible := false;
  //fb.SetFocus;
end;

procedure TBootloaderForm.ListBox1KeyDown(Sender: TObject; var Key: Word;
  Shift: TShiftState);
begin
  ListBox1.Visible := false;
end;

procedure TBootloaderForm.FormResize(Sender: TObject);
begin
  Myresize.Resize(BootloaderForm);
end;

function ResizeCallBack(Control: TControl; VFactor, HFactor: real; L, W, H, T, S, D1, D2: integer): boolean;
begin
  Result := false;
end;

procedure TBootloaderForm.FormCreate(Sender: TObject);
var IniFile: TSimpleInifile;
    W, H, L, T: word;
begin
  MyResize := TMyResize.Create(BootloaderForm, ResizeCallback, true); // also adapt fontsize

  Inifile := TSimpleIniFile.Create;
  W := IniFile.ReadInteger('Sizes', 'MainWidth', 461);
  H := IniFile.ReadInteger('Sizes', 'MainHeight', 506);
  L := IniFile.ReadInteger('Sizes', 'MainLeft', (Screen.Width - Width) div 2);
  T := IniFile.ReadInteger('Sizes', 'MainTop', (Screen.Height - Height) div 2);

  SetBounds(L, T, W, H);

  Application.ProcessMessages;

  if IniFile.ReadBool('Sizes', 'MainMaxiMised', false)
    then WindowState := wsMaximized
  else WindowState := wsNormal;

  Font.Size := IniFile.ReadInteger('Font', 'Main', 8);
  IniFile.Free;
  
end;

procedure TBootloaderForm.FormDestroy(Sender: TObject);
var
  IniFile: TSimpleIniFile;
begin
  MyResize.Free;

  IniFile := TSimpleIniFile.Create;

  if WindowState = wsMaximized
    then IniFile.WriteBool('Sizes', 'MainMaxiMised', true)
  else
  begin
    IniFile.WriteBool('Sizes', 'MainMaxiMised', false);
    IniFile.WriteInteger('Sizes', 'MainWidth', Width);
    IniFile.WriteInteger('Sizes', 'MainHeight', Height);
    IniFile.WriteInteger('Sizes', 'MainLeft', Left);
    IniFile.WriteInteger('Sizes', 'MainTop', Top);
  end;

  IniFile.WriteInteger('Font', 'Main', Font.Size);
  IniFile.Free;
end;

initialization
  ProgDir := ExtractFilepath(Application.Exename);
  if ProgDir[Length(ProgDir)] <> '\'
    then ProgDir := ProgDir + '\';
  ChDir(ProgDir);
end.
