UNIT BL_Vars;

INTERFACE
USES
  Forms,
  Types,
  SysUtils,
  HIDinterface;

TYPE
  TSetup                  = Record
                              Xb, Yb     : Integer;
                              Hexfilename_P18 : String[255];
                              Hexfilename_P24 : String[255];
                              Hexfilename_P32 : String[255];
                              Hexfilename_P16 : String[255];
                              VID         : Word;
                              PID         : Word;
                              DisableBL   : Boolean;
                            end;

CONST
  // command defintions
  STX                     = $0F;  // Start command
  cmdNON                  = 0;    // 'Idle'.
  cmdSYNC                 = 1;    // Synchronize with PC tool.
  cmdINFO                 = 2;    // Send bootloader info record.
  cmdBOOT                 = 3;    // Go to bootloader mode.
  cmdREBOOT               = 4;    // Restart MCU.
  cmdWRITE                = 11;   // Write to MCU flash.
  cmdREAD                 = 12;   // read from MCU flash.
  cmdERASE                = 21;   // Erase MCU flash.
  cmdClrBootFlag          = 22;   // Erase Boot flag.
  cmdSetBootFlag          = 23;   // Erase Boot flag.

  // verbose representation of command definitions
  csdNON                  = ' (Idle)';
  csdSYNC                 = ' (Sync)';
  csdINFO                 = ' (Info)';
  csdBOOT                 = ' (Boot)';
  csdREBOOT               = ' (Reboot)';
  csdWRITE                = ' (Write)';
  csdREAD                 = ' (Read)';
  csdERASE                = ' (Erase)';

  // MCU types
  mtPIC16                 = 1;
  mtPIC18                 = 2;
  mtPIC18FJ               = 3;
  mtPIC24                 = 4;
  mtDSPIC                 = 10;
  mtPIC32                 = 20;

  // WideString representation of mcu types
  stPIC16                 = 'PIC16';
  stPIC18                 = 'PIC18';
  stPIC18FJ               = 'PIC18FJ';
  stPIC24                 = 'PIC24';
  stDSPIC                 = 'DSPIC';
  stPIC32                 = 'PIC32';

  // bootloader record field definitions
  bifMCUTYPE              = 1; // MCU type/family.
  bifMCUID                = 2; // MCU ID number.
  bifERASEBLOCK           = 3; // MCU flash erase block size.
  bifWRITEBLOCK           = 4; // MCU flash write block size.
  bifBOOTREV               = 5; // Bootloader revision.
  bifBOOTSTART             = 6; // Bootloader start address.
  bifDEVDSC                = 7; // Device descriptor WideString.
  bifMCUSIZE               = 8; // MCU flash size

  MAX_String_FIELD_LENGTH = 20;
  
  HIDbufferSize           = 64;
  SettingsFilename        = 'PIC_Univ_Loader.dat';
  CRLF                    = #13#10;
  clOrange                = $007FFF;


TYPE
  TFlashImage             = Array[0..$7FFFF] of Byte;
  TBootImage              = Array[0..$2FFF] of Byte;

//  Bootloader info field types:
// Byte field (1 byte).
  TCharField = Record
    fFieldType: Byte;
    fValue    : Byte;
  end;

// Int field (2 bytes).
  TUIntField = Record
    fFieldType: Byte;
    intVal    : Word;
  end;

// Long field (4 bytes).
  TULongField = Record
    fFieldType: Byte;
    fValue    : DWord;
  end;

  TStringField = Record
    fFieldType: Byte;
    fValue    : String[MAX_String_FIELD_LENGTH];
  end;

// Bootloader info record (device specific information).
  TBootInfo = Record
    bSize       : Byte;
    bMcuType    : TCharField;
    ulMcuSize   : TULongField;
    uiEraseBlock: TUIntField;
    uiWriteBlock: TUIntField;
    uiBootRev   : TUIntField;
    ulBootStart : TULongField;
    sDevDsc     : TStringField;
  end;

VAR
  ProgDir                 : String   = '';
  DevList                 : THIDDeviceList;
  BootloaderActivated     : Boolean  = FALSE;
  CurrentDevice           : Integer  = -1;
  BootInfo                : TBootInfo;
  InBuf,
  OutBuf                  : THIDbuffer;
  State                   : Integer  = 0;
  NoUpdate                : Boolean  = FALSE;
  Setup                   : TSetup;
  TimStart                : Integer  = -1;
  ErrorCode               : Int64    = $7FBE0F;
  Event                   : Integer  = 0;
  FlashImage              : TFlashImage;
  BootImage               : TBootImage;
  EOFlash                 : Int64;
  EOBoot                  : Int64;
  InStr                   : String;
  DeviceConnected         : Boolean = FALSE;
//  Ports, oPorts           : i2cData;
  CurrPort                : Integer = -1;
  PortNr                  : Integer  = -1;
  A2ok                    : Boolean = FALSE;
  readyToFlash            : Byte    = 0;
  TargetDevice            : byte = 255;
  BootloadHeader          : THIDBuffer;
  BootStartZeroBased      : DWord;


Procedure InitVars;

IMPLEMENTATION

Procedure InitVars;
Begin
  BootloaderActivated := FALSE;
  CurrentDevice       := -1;
End;

INITIALIZATION
  ProgDir  := ExtractFilepath(Application.Exename);
  If ProgDir[Length(ProgDir)] <> '\'
    then ProgDir := ProgDir + '\';
  ChDir(ProgDir);
END.


