UNIT MG_IO;

{$IFDEF NoDebug}

{$O+} // Optimisation ON
{$D-} // Debug information OFF
{$I-} // I/O checking OFF
{$L-} // Local Symbols OFF
{$Q-} // Overflow Checking OFF
{$R-} // Range Checking OFF

{$ELSE}
{$O-} // Optimisation OFF
{$D+} // Debug information ON
{$I+} // I/O checking ON
{$L+} // Local Symbols ON
{$Q+} // Overflow Checking ON
{$R+} // Range Checking ON

{$ENDIF}

INTERFACE

uses HIDinterface{, Forms};

VAR
  TargetConnected   : Boolean    = FALSE;
  Ending            : Boolean    = FALSE;
  Target            : THIDdeviceInfo;

  NewDevices, OldDevices    : THIDDeviceList;
  NewNumDevices, OldNumDevices : byte;

Procedure SendData( Data : String);
Procedure FindDevices;  stdcall;

IMPLEMENTATION
USES
  BL_Vars,
  SysUtils;


CONST
  CRLF          = #13#10;

VAR
  Devices       : THIDDeviceList;


(*
Procedure ProcessData(Data : THIDbuffer);  stdcall;
Var
  InStr   : String;
  ByteNum : Byte;
Begin
  InStr   := '';
  ByteNum := 0;
  For ByteNum := 0 to 63 do
    InStr := InStr + CHR(Data[ByteNum]);
End;
*)

Procedure SendData( Data : String);
(*
  Send data over HID pipe:
  HID ALWAYS sends 64 bytes, so the output buffer must be
  cleared in between consecutive transfers to allow the slave
  to detect the end of data.
  In case string data is transferred character 0 flags the
  end of the data stream.
*)
Var
  OutBuff    : THIDbuffer;
  NrOfBytes   : Integer;
  PosInString : Integer;
  ByteNum     : Byte;

  Procedure ClearBuffer;
  Var
    Cnt : Byte;
  Begin
    For Cnt := 0 to SizeOf(OutBuff) - 1 do OutBuff[Cnt] := 0;
    ByteNum := 0;
  End;

Begin
  If NOT TargetConnected then Exit;
  If Ending then Exit;
  
  ClearBuffer;
  PosInString := 1;  // characters start at position 1 in a Pascal string
  NrOfBytes   := Length(Data);
//writeln('>', Data);
  If (NrOfBytes > 0) then
    While ((PosInString + ByteNum) <= NrOfBytes) do
      begin
        While (ByteNum < SizeOf(OutBuff)) AND ((PosInString + ByteNum) <= NrOfBytes) do
          begin
            OutBuff[ByteNum] := ORD(Data[PosInString + ByteNum]);
            INC(ByteNum);
          end;
        PosInString := PosInString + ByteNum;
        HIDwriteDevice(Target, OutBuff);  // send data to HID device
//        writeln(data);
        ClearBuffer;  // clear buffer between transfers to remove "old" data
      end;
End;

Procedure CtrlBootloader(Mode : Byte);
// Mode 0 : disable bootloader
//      1 : enable bootloader
//      2 : enablle bootloader and reset
Var
  Cmd : String;
Begin
//  If NOT Connected then Exit;
  Cmd := '         ';
  Cmd[ 1] := CHR($09);  // length
  Cmd[ 2] := CHR($00);
  Cmd[ 3] := CHR($01);  // user id
  Cmd[ 4] := CHR($01);  // group id: system
  Cmd[ 5] := CHR($06);  // upgrade
  Cmd[ 6] := CHR($03);  // set mode
  Cmd[ 7] := CHR($01);  // message number
  Cmd[ 8] := CHR(Mode); // set mode:
                        //  $00 - edisble bootloader
                        //  $01 - enable bootloader
                        //  $03 - enable bootloader and reset
  Cmd[ 9] := CHR($00);  // crc byte
  SendData(Cmd);
End;

Procedure FindDevices;  stdcall;
Var
  NumDevices, I : Byte;
  PID, VID: word;
begin

  VID := 0;
  PID := 0;
  HIDscanForDevices( Devices, NumDevices, VID, PID);
  If (NumDevices > 0) then
  begin
    For I := 0 to NumDevices - 1 do
      begin
        Target := Devices[I];
        TargetConnected := (Target.VID = Setup.VID) AND HIDopenDevice(Target);
        If TargetConnected then Break;
      end;
  end;
//  If TargetConnected then CtrlBootloader($03);
end;

INITIALIZATION
  OldNumDevices := 0;
  NewNumDevices := 0;
  HIDscanForDevices(OldDevices, OldNumDevices, 0, 0);
  USBsetEventHandler(@FindDevices);  // assign callback to detect USB activity
  //HIDSetEventHandler(@HIDReceive);
  HIDSetEventHandler(nil);
  //HidReceived := false;

  FINALIZATION
  //HIDSetEventHandler(nil);
  If TargetConnected then
    begin
      HIDcloseDevice(Target);
      USBsetEventHandler(Nil);
    end;
END.
